function [ samples,logLikePlot ] = DRIFT_MCMC(Y, missingEdges, numSamplesPerChain, burnIn, sampleGap, numChains, alpha, alpha_a, alpha_b, gamma, delta, WPriorSigma, initialNumFeatures)
%MCMC Sampler for DRIFT
%@author James Foulds, some code by Arthur Asuncion
%Note: I am storing the sampled edges in the Y matrix, so don't assume Y to
%stay unchanged or be ground truth.
%Note:  This code probably breaks group DRIFT!
%Note: assumes missingEdges doesn't contain duplicates
%@param Y a cell array of sociomatrices
%@param missingEdges a 2-d matrix where each row contains (graphIndex timestamp sender
%reciever) of a missing edge. (graphIndex is 1 if we only have one graph
%per timestep)
%
% AA: MODIFIED TO INCLUDE A CACHE TO MAKE IT RUN 10x FASTER

    numActors = size(Y{1}{1},1);
    T = length(Y{1}); %number of timesteps
    
    H_NT = sum(1./(1:numActors*T));% The T'th Harmonic number.
        
    wIsIdentity = false;
    isSymmetric = true; %Is W forced to be symmetric?
                        %This is currently hard-coded in computeMultiGraphLogLikeFast, so updating this here won't change it there!!!! FIXME
                        %Need to fix this if you want to set it to false currently.
                        %There are other changes by Arthur that need to be removed if we don't
                        %make W symmetric, e.g. he resymmetrizes when updating it or adding new features
                        %-Jimmy
                       
    samples = cell(1, numSamplesPerChain * numChains);
    
    % Hyperparameters.
    if nargin < 7
        alpha=3;
        alpha_a=3;
        alpha_b=1;
        gamma=3;
        delta=1;
        WPriorSigma=0.1;
        initialNumFeatures = 10;
    end
    
    % AA: Computing some statistics so I don't have to compute it in the loop
    Ydata.numGraphs = length(Y);  % bug here : size(Y,1)
    Ydata.numActors = numActors;
    Ydata.T = T;
    for tt=1:T
        Ydata.aggregate(tt).graph = zeros(numActors,numActors);
        % This aggregate graph per time slice comes in handy when computing
        % the likelihood quickly.
        for g = 1:Ydata.numGraphs
            Ydata.aggregate(tt).graph = Ydata.aggregate(tt).graph + Y{g}{tt};
        end
    end

    %sampling loop
    currentSample = 1;
    iter=1;
    for c = 1:numChains
        rand('seed',c);  % Set random seed for each new chain.
        [sample, Y, Ydata] = initialize(initialNumFeatures, alpha,gamma,delta, WPriorSigma, numActors, T, wIsIdentity, missingEdges, Y, Ydata, isSymmetric);
        
        % Initialize eps to correspond to density .1
        % TODO: set this using data.
        defaultDensity = .1;
        sample.eps = log(defaultDensity/(1-defaultDensity));
        
        burnInCount = 1;
        
        %fprintf('starting new chain\n');
        while true
            %fprintf('Iteration %d\n', iter);
            assert(size(sample.Z,3) == size(sample.a,1));
            
            %sample auxiliary slice variable s (for deciding on the
            %dimensionality)
            s = rand() .* min(sample.a); %semi-ordered stick breaking
            sample = extendRepresentation(sample, s, alpha, gamma, delta, WPriorSigma, wIsIdentity);
            numFeatures = size(sample.a,1);

            % AA: Here I create the likelihood cache matrices for WZ', ZWZ',
            % sigmoid(ZWZ'), log(sigmoid(ZWZ')), and log(1-sigmoid(ZWZ'))

            for tt=1:T
                % WZ' matrix:
                sample.cache(tt).wz = sample.W * (squeeze(sample.Z(tt,:,:)))';
                % ZWZ' matrix:
                sample.cache(tt).zwz = (squeeze(sample.Z(tt,:,:))) * sample.cache(tt).wz;
                % sigmoid(ZWZ') matrix:
                sample.cache(tt).szwz = 1 ./ (1 + exp(-(sample.cache(tt).zwz + sample.eps)));
                % log(sigmoid(ZWZ')) matrix:
                sample.cache(tt).logSZWZ = log(sample.cache(tt).szwz);
                % log(1-sigmoid(ZWZ')) matrix:
                sample.cache(tt).logOneMinusSZWZ = log(1-sample.cache(tt).szwz);
            end
            
            %sample Z's
            for i = 1:numActors
               for k = 1:numFeatures
                  % AA: This calls the fast version of the FB sampler
                  % (where I improved the log-likelihood calculation)
                  sample = forwardBackwardV2Fast(Y, Ydata, sample, i, k);
               end
            end

            sample = deleteEmptyColumns(sample);

            %sample alpha hyper-parameter
            sample = sampleAlpha( sample, alpha_a, alpha_b, H_NT);
            
            %sample a's, b's
            sample = sample_ab(sample, alpha, gamma, delta);
 
            % AA: Here I create the likelihood Cache again since columns
            % were deleted (this could actually be optimized since we could
            % just delete the corresponding columns in the caches as well)

            for tt=1:T
                % WZ' matrix:
                sample.cache(tt).wz = sample.W * (squeeze(sample.Z(tt,:,:)))';
                % ZWZ' matrix:
                sample.cache(tt).zwz = (squeeze(sample.Z(tt,:,:))) * sample.cache(tt).wz;
                % sigmoid(ZWZ') matrix:
                sample.cache(tt).szwz = 1 ./ (1 + exp(-(sample.cache(tt).zwz + sample.eps)));
                % log(sigmoid(ZWZ')) matrix:
                sample.cache(tt).logSZWZ = log(sample.cache(tt).szwz);
                % log(1-sigmoid(ZWZ')) matrix:
                sample.cache(tt).logOneMinusSZWZ = log(1-sample.cache(tt).szwz);
            end

            %sample intercept term eps

            for iii=1:5
                sample = sampleEps(sample, Y, Ydata, WPriorSigma);
            end

            % AA: Here I create the likelihood Cache again
            for tt=1:T
                % WZ' matrix:
                sample.cache(tt).wz = sample.W * (squeeze(sample.Z(tt,:,:)))';
                % ZWZ' matrix:
                sample.cache(tt).zwz = (squeeze(sample.Z(tt,:,:))) * sample.cache(tt).wz;
                % sigmoid(ZWZ') matrix:
                sample.cache(tt).szwz = 1 ./ (1 + exp(-(sample.cache(tt).zwz + sample.eps)));
                % log(sigmoid(ZWZ')) matrix:
                sample.cache(tt).logSZWZ = log(sample.cache(tt).szwz);
                % log(1-sigmoid(ZWZ')) matrix:
                sample.cache(tt).logOneMinusSZWZ = log(1-sample.cache(tt).szwz);
            end
            
            %sample the Ws using Metropolis-Hastings updates
            if ~wIsIdentity
                % AA: Look at sampleW function since I have modified it to 
                % speed it up as well
                % AA: I let it sample for 10 iterations
                for iterations = 1:10
                    sample = sampleW(sample, Y, Ydata, WPriorSigma);
                end
            end
            
            %perform Gibbs updates on the missing edges in Y
            [sample, Y, Ydata] = sampleMissingEdges(Y, Ydata, missingEdges, sample, isSymmetric);
                      
            % AA: Here I'm calling the fast loglikelihood function
            logLikelihood = DRIFTlogLikeFast(Y,Ydata, sample);

            
            %We don't want to count the missing edges in the loglikelihood,
            %so subtract these off from what we already computed.
            %TODO this could be made faster, eg taking advantage of caching
            %Note: assumes missingEdges doesn't contain duplicates!
            %This only matters for plotting the likelihood for convergence purposes.
            %Try putting this back if you need it, but I haven't tested it since
            %Arthur's caching code. -Jimmy
%             GRAPHINDEX = 1;
%             TIMESTAMP = 2;
%             SENDER = 3;
%             RECEIVER = 4;
%             for i = 1:size(missingEdges,1)
%                 missingEdges(i,:)
%                 size(sample.Z,1)
%                 size(sample.Z,2)
%                 size(sample.Z,3)
%                if Y{missingEdges(i,GRAPHINDEX)}{missingEdges(i,TIMESTAMP)}(missingEdges(i,SENDER), missingEdges(i,RECEIVER)) == 1
%                    logLikelihood = logLikelihood - log(1./(1 +   exp(-(squeeze(sample.Z(missingEdges(i,TIMESTAMP), missingEdges(i,SENDER),:))' * sample.W * squeeze(sample.Z(missingEdges(i,GRAPHINDEX), missingEdges(i,RECEIVER),:)) + sample.eps)))); %NOTE: If we introduce covariates we need to include them here!
%                else
%                    logLikelihood = logLikelihood - log(1-1./(1 + exp(-(squeeze(sample.Z(missingEdges(i,TIMESTAMP), missingEdges(i,SENDER),:))' * sample.W * squeeze(sample.Z(missingEdges(i,GRAPHINDEX), missingEdges(i,RECEIVER),:)) + sample.eps)))); %NOTE: If we introduce covariates we need to include them here!
%                end
%             end

            logLikePlot(iter) = logLikelihood;

            if burnInCount > burnIn && mod(iter - burnIn, sampleGap) == 0
                samples{currentSample} = sample;
                if mod(currentSample, numSamplesPerChain) == 0 %end of current chain
                    currentSample = currentSample + 1;
                    iter = iter + 1;
                    break;
                end
                currentSample = currentSample + 1;
            end
            iter = iter + 1  
            burnInCount = burnInCount + 1;


        end
    end
    %LL vs Iteration plot (optional)
    %figure();
    %plot(logLikePlot);
end

%Takes log probabilities as input, whose probabilities are assumed to sum
%to one, and computes the normalized probability of temp1.
function pr = normalizeLogProbs(temp1, temp0)
    a = max([temp1 temp0]);
    b = [exp(temp1 - a), exp(temp0 - a)];
    pr = b(1) ./ (b(1) + b(2));
end

%delete columns from W and Z for which there are no one entries
%(dishes that nobody is eating anymore)
function sample = deleteEmptyColumns(sample)
    i = 1;
    while true
       if i > size(sample.Z,3)
           break;
       end
       if size(sample.Z,3) == 2%1    %Setting the minimum number of features to 2 instead of 1, THIS MAY AVOID A BUG IN LL
           break; %don't want to end up with no features
       end
       if nnz(sample.Z(:,:,i)) == 0
           sample.Z(:,:,i) = [];
           sample.W(:,i) = [];
           sample.W(i,:) = [];
           sample.a(i) = [];
           sample.b(i) = [];

           i = i - 1;
       end
       i = i + 1;
    end

    assert(size(sample.W,1) == size(sample.W,2))
    assert(size(sample.W,1) == size(sample.Z,3))
    assert(size(sample.W,1) == size(sample.a,1))
end


%sample from a poisson distribution
%Knuth's algorithm from the wikipedia page
%Only use this if the statistics toolbox is not available.
function n = samplePoisson(lambda)
    
    L = exp(-lambda);
    k = 0;
    p = 1;
    while (true)
        k = k + 1;
        u = rand(1,1);
        p = p .* u;
        if (p <= L)
            n = k - 1;
            break;
        end
    end
end

%Only use this if the statistics toolbox is not available.
function x = sampleGaussian(mu, sigma)
  %first generate a standard normal number
  
  %using Box-Muller method (from Wikipedia)
  u = rand(1,1);
  v = rand(1,1);
  x = sqrt(-2 .* log(u)) .* cos(2 .* pi .* v);
  
  %using Irwin-Hall distribution approximation (from Wikipedia)
  %x = 0;
  %for i = 1:12
  %    x= x + rand(1,1);
  %end
  %x = x - 6;

  %get normal(mu,sigma) number
  x = mu + sigma .* x;
end

%Only use this if the statistics toolbox is not available.
function x = sampleExponential(lambda)
  x = rand(1,1);
  x = -log(x) ./ lambda;
end

%%
% Compute the (unnormalized) log posterior density for stick length a_k conditioned on 
% previous stick length a_k_minusone and the
% fact that all future chains for all actors consist entirely of zeros
% used for extending the dimensionality according to the slice variable
function [ logP ] = stickPosterior( a_k, params)
    if a_k < 0 ||  a_k > params.a_k_minusone
        logP = -inf;
        return
    end
    effectiveNumRows = params.numActors .* params.T;

    logP = (params.alpha-1) .* log(a_k) + effectiveNumRows * log(1-a_k);
    for t=1:effectiveNumRows
        logP = logP + params.alpha * (1-a_k).^t / t; %Equation 31 doesn't have this alpha, while equation 25 does.  Dilan thinks it's a typo
        %logP = logP + (1-a_k).^t / t;
    end
end


%count r_k, the number of ones that follow a zero, and m_k, the
%total number of zeros including the zero'th timestep and
%excluding the last timestep, and r_bk, m_bk, similar to the above
%but following a one.
function[r_k m_k r_bk m_bk] = countStickSufficientStatistics(Z,k)
   numActors = size(Z,2);
   T = size(Z,1);
   r_k = 0;
   m_k = numActors; %since the zero'th timestep is effectively zero, for each actor
   r_bk = 0;
   m_bk = 0;
   %TODO is there a better way to do this?
   %at the least we could cache these sufficient statistics
   previousWasZero = true;
   for i = 1:numActors
      for t = 1:T
          if Z(t,i,k)
              if previousWasZero
                    r_k = r_k + 1;
              else
                    r_bk = r_bk + 1;
              end
              if t < T
                    m_bk = m_bk + 1;
              end
              previousWasZero = false;
          else
              if t < T
                    m_k = m_k + 1;
              end
              previousWasZero = true;
          end
      end
      previousWasZero = true;
   end
end

function sample = extendRepresentation(sample, s, alpha, gamma, delta, WPriorSigma, wIsIdentity)
    %resize the a's, b',s Zs and Ws according to the new dimensionality given by s
    %a_k_minusone = sample.a(end); %this is for ordered stick breaking
    numActors = size(sample.Z,2);
    T = size(sample.Z,1); %number of timesteps
    numFeatures = size(sample.Z,3);
    
    a_k_minusone = min(sample.a);
    %a_k_afterLast = 0; %K^{\cross} in Dilan's paper. The probability for the feature after the ones we have nonzero entries for
    while a_k_minusone > s

        % Sample a new a and b.
        params.a_k_minusone = a_k_minusone;
        params.alpha = alpha;
        params.T = T;
        params.numActors = numActors;
        a_k = slicesample( 1, 20, @stickPosterior, rand() * a_k_minusone, 2, params);
        if a_k > s
            % Update data structures. (we leave W for afterwards)
            assert(size(sample.Z,3) == size(sample.a,1));
            sample.Z(:, :, end + 1) = 0; %TODO resize once only
            sample.a(end + 1) = a_k;
            sample.b(end + 1) = random('beta', gamma, delta);
            if size(sample.a,2) > 1
                %fixes a bug where it becomes a row vector instead of a
                %column vector
                sample.a = sample.a';
                sample.b = sample.b';
            end

            assert(size(sample.Z,3) == size(sample.a,1));
        else
            %a_k_afterLast = a_k;
        end
        a_k_minusone = a_k;
    end
    
    
    %fprintf('num features: %d.\n', size(sample.a,1));
    
    
    assert(size(sample.Z,3) == size(sample.a,1));
    %resize W
    numNewDishes = size(sample.Z,3) - size(sample.W,1);
    if wIsIdentity
        sample.W = eye(numFeatures + numNewDishes);
    else
        if numNewDishes == 0
            return;
        end
        WPrime = zeros(numFeatures + numNewDishes);
        WPrime(1:size(sample.W,1), 1:size(sample.W,2)) = sample.W;
        sample.W = WPrime;
        for j = (numFeatures + 1):(numFeatures + numNewDishes)
            for k = 1:(numFeatures + numNewDishes)
                %WPrime(j,k) = sampleGaussian(0, WPriorSigma);
                sample.W(j,k) = random('norm', 0, WPriorSigma); %need the statistics toolbox for this!!
            end
        end
        for j =1:numFeatures
            for k= (numFeatures + 1):(numFeatures + numNewDishes)
                %WPrime(j,k) = sampleGaussian(0, WPriorSigma);
                sample.W(j,k) = random('norm', 0, WPriorSigma); %need the statistics toolbox for this!!
            end
        end
        
        %AA: Hack: Make sure W is symmetric (can make this faster)
        for j=2:(numFeatures+numNewDishes)
            for k=1:(j-1)
                sample.W(j,k) = sample.W(k,j);
            end
        end
        
    end
    numFeatures = size(sample.a,1);
    assert(size(sample.W,1) == size(sample.W,2))
    assert(size(sample.W,1) == size(sample.Z,3))
    assert(size(sample.W,1) == numFeatures)

end

function sample = sample_ab(sample, alpha, gamma, delta)
    numActors = size(sample.Z,2);
    T = size(sample.Z,1); %number of timesteps
    numFeatures = size(sample.Z,3);
    %a_k_minusone = 1.0;
    for k = 1:numFeatures        
       %semi-ordered stick breaking
       [r_k m_k r_bk m_bk] = countStickSufficientStatistics(sample.Z,k);
       if r_k ~= 0
           sample.a(k) = random('beta', r_k, 1 + m_k - r_k); %equation 30 in Teh et al.
       else
           %since we deleted empty columns, but don't let the array become empty, the only way this can
           %happen is if there is only one (empty) column left
           %assert(size(sample.a,1) == 1);
           assert(size(sample.a,1) == 2);
           %If I didn't delete empty columns I need to use Equation 31
           %instead.
           params.a_k_minusone = alpha;
           params.alpha = alpha;
           params.T = T;
           params.numActors = numActors;
           sample.a(k) = slicesample( 1, 20, @stickPosterior, rand() * alpha, 2, params);
       end

       sample.b(k) = random('beta', gamma + r_bk, delta + m_bk - r_bk);
    end
end

% AA: This function now takes in Ydata because the aggregated graphs are
% needed when computing the loglikelihood quickly
function sample = sampleW(sample, Y,Ydata, WPriorSigma)
    numFeatures = size(sample.Z,3);
    for j = 1:numFeatures
        for k =j:numFeatures  % AA: Used to be 1:numFeatures, but we assume symmetry
            %WPrime(j,k) = sampleGaussian(0, WPriorSigma);

            %currentLLslow = DRIFTlogLike(Y, sample);
            % AA: optimized function
            currentLL = DRIFTlogLikeFast(Y,Ydata, sample);
            
            %assert(abs(currentLL - currentLLslow) < 0.001);
            
            currentWjk = sample.W(j,k);
            currentPrior = log(normpdf(sample.W(j,k),0,WPriorSigma));
            %sample.W(j,k) = random('norm', 0, WPriorSigma); %sample from the prior
            oldEntry = sample.W(j,k);
            sample.W(j,k) = random('norm', sample.W(j,k), WPriorSigma); %sample from Gaussian centered at current location
            % AA: Make sure symmetry is preserved
            sample.W(k,j) = sample.W(j,k);
            sample = updateCacheW(sample,j,k, oldEntry);
            
            %proposalLLslow = DRIFTlogLike(Y, sample);
            % AA: optimized function
            proposalLL = DRIFTlogLikeFast(Y,Ydata, sample);
            
            %assert(abs(proposalLL - proposalLLslow) < 0.001);
            
            proposalPrior = log(normpdf(sample.W(j,k),0,WPriorSigma));
            %acceptanceRatio = exp(proposalLL - currentLL); %use this when sampling from the prior
            acceptanceRatio = exp(proposalLL - currentLL + proposalPrior - currentPrior);
            if rand() < acceptanceRatio
                %accept the proposal
                %We don't need to do anything as we already changed
                %it
            else
                %go back to the old value
                oldEntry = sample.W(j,k);
                sample.W(j,k) = currentWjk;
                % Make sure symmetry is preserved
                sample.W(k,j) = sample.W(j,k);
                sample = updateCacheW(sample,j,k, oldEntry);
            end
        end
    end
    
    %AA: Hack: Make sure W is symmetric (can make this faster)
    for j=2:numFeatures
        for k=1:(j-1)
           sample.W(j,k) = sample.W(k,j);
        end
    end
    
end


% AA: This is a new helper function that updates my cached matrix
% statistics after changing one W entry (called by sampleW)
function [sample] = updateCacheW(sample,j,k, oldEntry)
    % Updates the cache statistics, given that W(j,k) was changed
    % (where the original entry was oldEntry)

    T = size(sample.Z,1);

    for t=1:T % Have to change cache for each time step

        % WZ cache changes in 1 (or 2) different rows.  The deltas are the
        % values that these rows change by.  Note if j==k, one only has to
        % change one row.
        delta1 = (-oldEntry + sample.W(j,k)).* (squeeze(sample.Z(t,:,k)));
        if (j~=k)
            delta2 = (-oldEntry + sample.W(j,k)).* (squeeze(sample.Z(t,:,j)));
        end

        sample.cache(t).wz(j,:) = sample.cache(t).wz(j,:) + delta1;
        if (j~=k)
            sample.cache(t).wz(k,:) = sample.cache(t).wz(k,:) + delta2;
        end

        % One has to modify the WZ cache as well
        if (j~=k)
            sample.cache(t).zwz = sample.cache(t).zwz + squeeze(sample.Z(t,:,[j,k])) * [delta1; delta2];
        else
            sample.cache(t).zwz = sample.cache(t).zwz + squeeze(sample.Z(t,:,[j]))' * [delta1];
        end

        % Now need to take sigmoid of changed entries to update SZWZ.
        % Likewise, modify logSZWZ and logOneMinusSZWZ
        sample.cache(t).szwz = 1 ./ (1 + exp(-(sample.cache(t).zwz + sample.eps)));
        sample.cache(t).logSZWZ = log(sample.cache(t).szwz);
        sample.cache(t).logOneMinusSZWZ = log(1-sample.cache(t).szwz);

    end

end

function [sample, Y, Ydata] = initialize(numFeatures, alpha,gamma,delta, WPriorSigma, numActors, T, wIsIdentity, missingEdges, Y, Ydata, isSymmetric)
    %sample a, b from their priors
    sample.a = zeros(numFeatures,1);
    sample.b = zeros(numFeatures,1);
    for k = 1:numFeatures
        if k == 1
            sample.a(1) = random('beta', alpha, 1);
        else
            stickRatio = random('beta', alpha, 1);
            sample.a(k) = sample.a(k-1) .* stickRatio;
        end
        sample.b(k) = random('beta', gamma, delta);
    end
    
    %sample Z from its prior
    sample.Z = zeros(T, numActors,numFeatures);
    for i = 1:numActors
        for j=1:T
           for k = 1:numFeatures
               if j == 1 || sample.Z(j - 1, i, k) == 0
                    pr_featureOn = sample.a(k);
               else pr_featureOn = sample.b(k);
               end
               sample.Z(j, i, k) = (rand() < pr_featureOn);
           end
        end
    end
    
    %sample W from its prior
    if wIsIdentity
        sample.W = eye(numFeatures);
    else
        sample.W = random('norm', 0, WPriorSigma,numFeatures,numFeatures); %need the statistics toolbox for this
    end
    %If no statistics toolbox:
%     W = zeros(numFeatures, numFeatures);
%     for j = 1:numFeatures
%         for k = 1:numFeatures
%             W(j,k) = sampleGaussian(0, WPriorSigma);
%         end
%     end
    
    %AA: Hack: Make sure W is symmetric (can make this faster)
    for j=2:numFeatures
        for k=1:(j-1)
           sample.W(j,k) = sample.W(k,j);
        end
    end

    sample.eps = 0;
    sample.missingEdgeSamples = zeros(size(missingEdges,1),1);
    %randomly initialize missing edges
    GRAPHINDEX = 1;
    TIMESTAMP = 2;
    SENDER = 3;
    RECEIVER = 4;
    for i = 1:size(missingEdges,1)
       prEdge = 1./(1 + exp(-(squeeze(sample.Z(missingEdges(i,TIMESTAMP), missingEdges(i,SENDER),:))' * sample.W * squeeze(sample.Z(missingEdges(i,TIMESTAMP), missingEdges(i,RECEIVER),:)) + sample.eps))); %NOTE: If we introduce covariates we need to include them here!!!!!!! TODO!!
       sample.missingEdgeSamples(i) = rand() < prEdge;
       Y{missingEdges(i,GRAPHINDEX)}{missingEdges(i,TIMESTAMP)}(missingEdges(i,SENDER),missingEdges(i,RECEIVER)) = sample.missingEdgeSamples(i);
       if isSymmetric
           Y{missingEdges(i,GRAPHINDEX)}{missingEdges(i,TIMESTAMP)}(missingEdges(i,RECEIVER),missingEdges(i,SENDER)) = Y{missingEdges(i,GRAPHINDEX)}{missingEdges(i,TIMESTAMP)}(missingEdges(i,SENDER),missingEdges(i,RECEIVER));
       end
    end
end

function [sample, Y, Ydata] = sampleMissingEdges(Y, Ydata, missingEdges, sample, isSymmetric)
%sample missing edges given our current parameters
    GRAPHINDEX = 1;
    TIMESTAMP = 2;
    SENDER = 3;
    RECEIVER = 4;
    for i = 1:size(missingEdges,1)
       %prEdge = 1./(1 + exp(-(squeeze(sample.Z(missingEdges(i,TIMESTAMP), missingEdges(i,SENDER),:))' * sample.W * squeeze(sample.Z(missingEdges(i,TIMESTAMP), missingEdges(i,RECEIVER),:)) + sample.eps))); %NOTE: If we introduce covariates we need to include them here!!!!!!! TODO!!
       prEdge = sample.cache(missingEdges(i,TIMESTAMP)).szwz(missingEdges(i,SENDER), missingEdges(i,RECEIVER));
       sample.missingEdgeSamples(i) = rand() < prEdge;
       Y{missingEdges(i,GRAPHINDEX)}{missingEdges(i,TIMESTAMP)}(missingEdges(i,SENDER),missingEdges(i,RECEIVER)) = sample.missingEdgeSamples(i);
       if isSymmetric
           Y{missingEdges(i,GRAPHINDEX)}{missingEdges(i,TIMESTAMP)}(missingEdges(i,RECEIVER),missingEdges(i,SENDER)) = Y{missingEdges(i,GRAPHINDEX)}{missingEdges(i,TIMESTAMP)}(missingEdges(i,SENDER),missingEdges(i,RECEIVER));
       end
    end
    
    %we changed Y, so we need to update our cache of the sum of the Y edges
    %used for computing the likelihood.
    T = length(Y{1}); %number of timesteps
    numActors = size(sample.Z,2);
    for tt=1:T
        Ydata.aggregate(tt).graph = zeros(numActors,numActors);
        for g = 1:Ydata.numGraphs
            Ydata.aggregate(tt).graph = Ydata.aggregate(tt).graph + Y{g}{tt};
        end
    end
end


function [ sample ] = sampleAlpha( sample, alpha_a, alpha_b, H_NT)
%sample the hyper-parameter alpha
    numFeatures = size(sample.Z,3); %TODO this is wrong if the number of features dropped below 2, since we
    %didn't delete those columns.

    sample.alpha = random('Gamma', numFeatures + alpha_a, 1./(H_NT + alpha_b));
end


function [sample] = sampleEps(sample, Y, Ydata, WPriorSigma)
%Sample the eps intercept term

    currentLL = DRIFTlogLikeFast(Y, Ydata, sample);

    currentEps = sample.eps;
    currentPrior = log(normpdf(sample.eps,0,WPriorSigma));
    sample.eps = random('norm', sample.eps, WPriorSigma); %sample from Gaussian centered at current location


    for tt=1:Ydata.T
        % sigmoid(ZWZ') matrix:
        sample.cache(tt).szwz = 1 ./ (1 + exp(-(sample.cache(tt).zwz + sample.eps)));
        % log(sigmoid(ZWZ')) matrix:
        sample.cache(tt).logSZWZ = log(sample.cache(tt).szwz);
        % log(1-sigmoid(ZWZ')) matrix:
        sample.cache(tt).logOneMinusSZWZ = log(1-sample.cache(tt).szwz);
    end

    proposalLL = DRIFTlogLikeFast(Y, Ydata, sample);
    
    proposalPrior = log(normpdf(sample.eps,0,WPriorSigma));
    acceptanceRatio = exp(proposalLL - currentLL + proposalPrior - currentPrior);
    if rand() < acceptanceRatio
        %accept the proposal
        %We don't need to do anything as we already changed
        %it
    else
        sample.eps = currentEps;
        
        for tt=1:Ydata.T
            % sigmoid(ZWZ') matrix:
            sample.cache(tt).szwz = 1 ./ (1 + exp(-(sample.cache(tt).zwz + sample.eps)));
            % log(sigmoid(ZWZ')) matrix:
            sample.cache(tt).logSZWZ = log(sample.cache(tt).szwz);
            % log(1-sigmoid(ZWZ')) matrix:
            sample.cache(tt).logOneMinusSZWZ = log(1-sample.cache(tt).szwz);
        end
        
    end
end




