function [ ll, incrementedSamples] = logLikeForFutureTimesteps( model_samples, WPriorSigma, samplesPerSample, Y_future, extendRepresentation)
%Sample the Zs for the next timestep from each of our MCMC samples of the
%model (without observing the data at that timestep Y_next), and compute the
%expected log-likelihood of the observed data.

    T_future = size(Y_future{1},1);
    sampleLogProbs = zeros(length(model_samples) .* samplesPerSample,1);
    slp_index = 1;
    incrementedSamples = cell(1);
    for s = 1:length(model_samples)
       for s2 = 1:samplesPerSample
           sample = model_samples{s};
           sample.Z = increment_Z(sample, T_future, extendRepresentation);
           if size(sample.Z,3) > size(sample.W,1)
               sample.W = extendW(sample.W, sample.Z, WPriorSigma, false);
           end
           assert(size(sample.Z,3) == size(sample.W,1));
           sampleLogProbs(slp_index) = DRIFTlogLike(Y_future, sample);
           incrementedSamples{slp_index} = sample;
           slp_index = slp_index + 1;
       end
    end
    ll = logOfSum(sampleLogProbs) - (log(length(model_samples)) + log(samplesPerSample));
end

function Z = increment_Z(sample, T_future, extendRepresentation)
%sample the value of Z in the next T_future timesteps
    T = size(sample.Z,1);
    numActors = size(sample.Z,2);
    numFeatures = size(sample.Z,3);
    Z = zeros(T_future, numActors, numFeatures);
    for t = 1:T_future
        for i = 1:numActors
           for k = 1:numFeatures
               pr_featureOn = 0;
               if (t == 1 && sample.Z(T,i, k) == 0) || (t > 1 && Z(t-1,i,k) == 0)
                   pr_featureOn = sample.a(k);
               else pr_featureOn = sample.b(k);
               end
               Z(t,i, k) = rand() < pr_featureOn;
           end
        end
    end
    
    if ~extendRepresentation
        return;
    end
    
    %sample unobserved features
    s = rand() .* min(sample.a); %semi-ordered stick breaking
    
    a_k_minusone = min(sample.a);
    while a_k_minusone > s

        % Sample a new a and b.
        params.a_k_minusone = a_k_minusone;
        params.alpha = sample.alpha;
        params.T = T;
        params.numActors = numActors;
        a_k = slicesample( 1, 20, @stickPosterior, rand() * a_k_minusone, 2, params);
        if a_k > s
            % Update data structures. (we leave W for afterwards)
            assert(size(Z,3) == size(sample.a,1));
            Z(:, :, end + 1) = 0;
            sample.a(end + 1) = a_k;
            %sample.b(end + 1) = random('beta', gamma, delta);
            if size(sample.a,2) > 1
                %fixes a bug where it becomes a row vector instead of a
                %column vector
                sample.a = sample.a';
                %sample.b = sample.b';
            end

            assert(size(Z,3) == size(sample.a,1));
        else
            %a_k_afterLast = a_k;
        end
        a_k_minusone = a_k;
    end
    %Z values for new features
    for t = 1:T_future
        for i = 1:numActors
           for k = numFeatures+1:size(sample.a,1);
               pr_featureOn = 0;
               if (t == 1) || (t > 1 && Z(t-1,i,k) == 0)
                   pr_featureOn = sample.a(k);
               else pr_featureOn = sample.b(k);
               end
               Z(t, i, k) = rand() < pr_featureOn;
           end
        end
    end
    
end

function W = extendW(W, Z, WPriorSigma, wIsIdentity)
%We added some new features to Z, so add their entries for W
%(sampled from the prior)
    numFeatures = size(W,1);
    numNewDishes = size(Z,3) - numFeatures;
    if wIsIdentity
        W = eye(numFeatures + numNewDishes);
    else
        if numNewDishes == 0
            return;
        end
        WPrime = zeros(numFeatures + numNewDishes);
        WPrime(1:size(W,1), 1:size(W,2)) = W;
        W = WPrime;
        for j = (numFeatures + 1):(numFeatures + numNewDishes)
            for k = 1:(numFeatures + numNewDishes)
                %WPrime(j,k) = sampleGaussian(0, WPriorSigma);
                W(j,k) = random('norm', 0, WPriorSigma); %need the statistics toolbox for this!!
            end
        end
        for j =1:numFeatures
            for k= (numFeatures + 1):(numFeatures + numNewDishes)
                %WPrime(j,k) = sampleGaussian(0, WPriorSigma);
                W(j,k) = random('norm', 0, WPriorSigma); %need the statistics toolbox for this!!
            end
        end
    end
    assert(size(W,1) == size(W,2))
    assert(size(W,1) == size(Z,3))
end


%%
% Compute the (unnormalized) log posterior density for stick length a_k conditioned on 
% previous stick length a_k_minusone and the
% fact that all future chains for all actors consist entirely of zeros
% used for extending the dimensionality according to the slice variable
function [ logP ] = stickPosterior( a_k, params)
    if a_k < 0 ||  a_k > params.a_k_minusone
        logP = -inf;
        return
    end
    effectiveNumRows = params.numActors .* params.T; 
    
    logP = (params.alpha-1) .* log(a_k) + effectiveNumRows * log(1-a_k);
    for t=1:effectiveNumRows
        logP = logP + params.alpha * (1-a_k).^t / t; %Equation 31 doesn't have this alpha, while equation 25 does.  Dilan thinks it's a typo
        %logP = logP + (1-a_k).^t / t;
    end
end




